import pdb
import time
from itertools import chain
from itertools import islice
import h5py
import mmh3
import numpy as np
from nltk.tokenize import sent_tokenize
from numpy.linalg import norm
import torch
from torch import nn
from torch import optim
from text_embedding.features import split_on_punctuation
from text_embedding.vectors import vocab2mat
from text_embedding.documents import evaluate
from text_embedding.documents import write
from UnFURL.utils import FLOAT


LINT = np.int64
EOS = '<EOS>'
UNK = '<UNK>'
PRINTERVAL = 1000
TASKS = ['mr', 'cr', 'subj', 'mpqa', 'trec', 'sst', 'sst_fine', 'imdb']


# NOTE: to convert embedding text file to HDF5 use function 'text_embedding.vectors.text2hdf5'
def word_embedding(vectorfile=None, vocabfile=None, min_count=10, dimension=300, bins=2**20, seed=0, **kwargs): 
    '''constructs nn.Embedding object from a given vocabulary or pretrained vectors
    Args:
        vectorfile: embedding file in HDF5 format with keys 'words' and 'vectors'
        vocabfile: file containing two space-separated columns for words and their corpus counts
        min_count: minimum number of times word appears in corpus ; ignored if not vectorfile is None
        dimension: embedding dimension ; ignored if not vectorfile is None
        bins: number of hashing bins ; ignored if not vectorfile is None or not vocabfile is None
        seed: hashing seed ; ignored if not vectorfile is None or not vocabfile is None
        kwargs: passed to nn.Embedding.from_pretrained
    Returns:
        torch.nn.Embedding object, tokenization function
    '''

    vocab = [EOS, UNK]

    if vectorfile is None:
        if vocabfile is None:
            vocab = bins
        else:
            try:
                with open(vocabfile, 'r') as f:
                    vocab.extend(word for word, count in (line.split() for line in f) if int(count) >= min_count)
            except ValueError:
                with open(vocabfile, 'r') as f:
                    vocab.extend(line.strip() for line in f)
        embeddings = vocab2mat(vocab, random=(-0.1, 0.1), dimension=dimension)

    else:
        f = h5py.File(vectorfile)
        words, vectors = np.array(f['words']), np.array(f['vectors'])
        words = np.array([word.decode('utf-8') for word in words])
        f.close()
        try:
            with open(vocabfile, 'r') as f:
                wordset = {line.split()[0] for line in f}
            select = np.fromiter((word in wordset for word in words), dtype=np.bool, count=words.shape[0])
            vocab.extend(words[select])
            vectors = vectors[select]
        except TypeError:
            vocab.extend(words)
        bound = np.std(vectors) / 4.0 / np.sqrt(3.0)
        embeddings = np.vstack((np.random.uniform(-bound, bound, size=(2, vectors.shape[1])).astype(FLOAT), vectors))

    if type(vocab) == int:
        str2idx = lambda token: LINT((mmh3.hash(token, seed) % (bins-1)) + 1)
    else:
        _str2idx = dict((word, LINT(i)) for i, word in enumerate(vocab))
        unk = _str2idx[UNK]
        str2idx = lambda token: _str2idx.get(token, unk)
    embeddings[0] = FLOAT(0.0)
    return nn.Embedding.from_pretrained(torch.from_numpy(embeddings), **kwargs), str2idx


class Encoder(nn.Module):
    '''CBOW Encoder'''

    def __init__(self, input_dimension=300, max_length=100, **kwargs):
        '''
        Args:
            input_dimension: word embedding dimension (also passed to word_embedding)
            max_length: maximum sentence length
            kwargs: passed to word_embedding
        '''

        super().__init__()
        self.embedding, self.vocab = word_embedding(dimension=input_dimension, **kwargs)
        self.max_length = max_length

    def str2long(self, texts, tokenize=lambda text: split_on_punctuation(text.lower()), device=torch.device('cuda')):
        '''encodes text strings as LongTensors
        Args:
            texts: list of strings
            tokenize: function taking a string as input and returning an iterable of strings as output
            device: Torch device
        Returns:
            LongTensor of shape (len(texts), max_length) on device
        '''

        vocab, max_length = self.vocab, self.max_length
        array = np.zeros((len(texts), max_length), dtype=LINT)
        for i, text in enumerate(list(islice(tokenize(text), max_length-1)) for text in texts):
            length = len(text)
            array[i,:length] = np.fromiter((vocab(token) for token in text), LINT, length)
        return torch.LongTensor(array).to(device)
        
    def init_params(self):
        
        return self

    def forward(self, input):
        '''
        Args: 
            input: LongTensor of word ids
        Returns:
            FloatTensor of CBOW sentence representations
        '''

        return self.embedding(input).sum(1) / torch.clamp((input>0).sum(1), 1).float()[:,None]


class TextEmbedding(nn.Module):
    '''class for text representation from an encoder
    '''

    def __init__(self, encoder=Encoder, context=True, **kwargs):
        '''
        Args:
            encoder: constructor of object extending Encoder that takes in raw input and outputs sentence representations
            context: bool to have separate context representation
            kwargs: passed to encoder
        '''

        super().__init__()           
        self.encoder = encoder(**kwargs).init_params()
        if context:
            self.context_encoder = encoder(**kwargs).init_params()
        self.context = context
    
    def str2float(self, texts, **kwargs):
        '''encodes text strings as FloatTensors
        Args:
            texts: list of strings
            kwargs: passed to self.encoder.str2long
        Returns:
            FloatTensor of sentence representations
        '''

        input = self.encoder.str2long(texts, **kwargs)
        if self.context:
            rep = self.encoder(input)
            context = self.context_encoder(input)
            if self.training:
                return rep, context
            return torch.cat([rep, context], 1)
        return self.encoder(input)        


    def represent(self, texts, batch_size=100, sum_of_sents=False, **kwargs):
        '''represents text strings as numpy arrays
        Args:
            texts: list of strings of words in a sentence 
            batch_size: number of texts to process at a time
            sum_of_sents: represent each text as the sum of the representations of its constituent sentences
            kwargs: passed to self.encoder.str2long
        Returns:
            numpy array of text representations
        '''

        if sum_of_sents:
            return np.vstack(np.sum(self.represent(sent_tokenize(text), batch_size=batch_size, sum_of_sents=False, **kwargs), axis=0) for text in texts)

        self.eval()
        iterator = iter(texts)
        batches = (list(islice(iterator, batch_size)) for i in range(0, len(texts), batch_size))
        output = np.vstack(self.str2float(batch, **kwargs).cpu().detach().numpy() for batch in batches)
        self.train()
        return output


def evaluate_model(model, tasks=['sst'], mean_clf=False, **kwargs):
    '''evaluates sentence encoder model on given tasks
    Args:
        model: TextEmbedding object
        tasks: list of strings (must be function names in text_embedding.documents)
        mean_clf: use mean classifier instead of logit
        kwargs: passed to model.represent
    Returns:
        None
    '''
    
    for task in tasks:
        evaluate(task, lambda sents: model.represent(sents, sum_of_sents=task=='imdb', **kwargs), invariant=True, verbose=True, n_jobs=1, mean_clf=mean_clf)


def train_model(model, criterion, corpus, epochs=1, opt=optim.Adam, clip=0.0, batch_size=100, verbose=False, device=torch.device('cuda'), eval_freq=0, eval_tasks=['sst'], save_freq=0, save_loc='temp', debug=False, **kwargs):
    '''trains sentence encoder model on text corpus
    Args:
        model: TextEmbedding object
        corpus: text corpus with one example on each line
        epochs: number of training epochs
        opt: constructor of object extending torch.nn.optimizer.Optimizer
        clip: norm of gradient clipping ; if 0.0 does not clip gradients
        batch_size: number of lines to process at a time
        verbose: display output
        device: Torch device
        eval_freq: frequency of evaluation
        eval_tasks: list of strings (must be function names in text_embedding.documents)
        save_freq: frequency of saving
        save_loc: save location
        debug: stop execution and enter debug mode at intermediate stages
        kwargs: passed to opt
    Returns:
        None
    '''

    model.to(device)
    criterion.to(device)
    optimizer = opt(filter(lambda param: param.requires_grad, chain(model.parameters(), criterion.parameters())), **kwargs)

    if eval_freq:
        evaluate_model(model, eval_tasks, batch_size=batch_size, device=device)

    for epoch in range(epochs):
        with open(corpus, 'r') as f:

            t, total_loss, total_lines, _lines = time.time(), 0.0, 0, 0
            while True:

                lines = list(islice(f, batch_size))
                if lines:
                    loss = criterion(model.str2float(lines, tokenize=str.split, device=device))
                    optimizer.zero_grad()
                    loss.backward()
                    if clip:
                        nn.utils.clip_grad_norm_(model.parameters(), clip)
                    optimizer.step()
                    batch_lines = len(lines)
                    total_loss += batch_lines*float(loss)
                    total_lines += batch_lines
                    if verbose and total_lines%PRINTERVAL < _lines+batch_lines:
                        write('\r'+60*' ')
                        write('\rProcessed '+str(total_lines)+' Lines')
                        write(' at '+str(round((time.time()-t)/total_lines, 4))+' Sec/Line.')
                        write(' Loss='+str(round(total_loss/total_lines, 4)))
                        _lines = total_lines%PRINTERVAL
                    else:
                        _lines += len(lines)
                else:
                    write('\r'+60*' ')
                    write('\rCompleted Epoch '+str(epoch+1))
                    write(' in '+str(time.time()-t)+' Seconds.')
                    write(' Loss='+str(total_loss/total_lines)+'\n')
                    if debug:
                        pdb.set_trace()
                    break
                    
        if eval_freq and epoch % eval_freq == eval_freq-1:
            evaluate_model(model, eval_tasks, batch_size=batch_size, device=device)
        if save_freq and epoch % save_freq == save_freq-1:
            save_str = save_loc + "_e{}".format(epoch)
            torch.save(model.state_dict(), save_str)
