import pdb
import numpy as np
import torch
from torch import nn
from torch.nn import functional
from torch.nn.modules import loss
from text_embedding.documents import write
from UnFURL.utils import FLOAT
from UnFURL.utils import print_dict


# TODO: test correctness
class PSD(nn.Linear):
    '''Linear layer with PSD matrix'''

    def __init__(self, dimension, **kwargs):
        '''
        Args:
            dimension: dimension of input representation
            kwargs: passed to nn.Linear
        '''

        super().__init__(dimension, dimension, **kwargs)
        self._forward = super().forward

    def forward(self, input):

        return self._forward(functional.linear(input, torch.t(self.weight)))


# TODO: test correctness
class Diagonal(nn.Linear):
    '''Diagonal matrix layer'''

    def __init__(self, dimension, **kwargs):
        '''Args:
            dimension: dimension of input representation
            bias: train bias
        '''

        super().__init__(dimension, 1, **kwargs)

    def forward(self, input):

        if self.bias is None:
            return self.weight * input
        return self.weight * input + self.bias


# TODO: test correctness
class Quadratic(nn.Bilinear):
    '''Quadratic scalar function of the input'''

    def __init__(self, dimension, **kwargs):
        '''
        Args:
            dimension: dimension of input representation
            kwargs: passed to nn.Bilinear
        '''

        super().__init__(dimension, dimension, 1, **kwargs)
        self._forward = super().forward

    def forward(self, input):

        return self._forward(input, input)


class UnFURL(nn.Module):
    '''Unsupervised loss function
    '''

    def __init__(self, context_size=3, max_batch_size=1000, objective='nce', transform=None, bias=None, dimension=None):
        '''
        Args:
            context_size: number of context points surrounding a data point (including point itself)
            max_batch_size: size of largest training batch to be used
            transform: 'diagonal', 'linear', or 'psd' ; if None uses identity ; if not None dimension must also be set
            bias: 'linear' or 'quadratic' ; if None uses nothing ; if not None dimension must also be set
            dimension: representation dimension ; only needs to be set if transform or bias is set
        '''

        super().__init__()

        target = np.zeros((max_batch_size, max_batch_size), dtype=FLOAT)
        for i in range(1, context_size-1):
            target += np.eye(max_batch_size, k=-i, dtype=FLOAT)
            target += np.eye(max_batch_size, k=i, dtype=FLOAT)
        target_sum = np.sum(target, axis=1, keepdims=True)
        target = target / target_sum
        self.register_buffer('target', torch.from_numpy(target / np.sum(target, axis=1, keepdims=True)))

        if transform is None:
            self.transform = None
        elif transform == 'diagonal':
            self.transform = Diagonal(dimension, bias=False)
        elif transform == 'linear':
            self.transform = nn.Linear(dimension, dimension, bias=False)
        elif transform == 'psd':
            self.transform = PSD(dimension, bias=False)
        else:
            raise(NotImplementedError)

        if bias is None:
            self.bias = None
        elif bias == 'linear':
            self.bias = nn.Linear(dimension, 1, bias=True)
        elif bias == 'quadratic':
            self.bias = Quadratic(dimension, bias=True)
        else:
            raise(NotImplementedError)

        if objective.lower() == 'nce':
            self.logsigmoid = nn.LogSigmoid()
            self.C = nn.Parameter(torch.FloatTensor([0.0]))
            self.objective = self.nce
        else:
            self.logsoftmax = nn.LogSoftmax(1)
            self.objective = self.cross_entropy

    def nce(self, pred):
        '''
        Args:
            pred: FloatTensor of inner product (plus bias) between representations of a batch.
        Returns:
            unsupervised noise contrastive estimation loss
        '''

        batch_size = pred.shape[0]
        targ = self.target[:batch_size,:batch_size]
        n_pos = (targ > 0).sum().float()

        # TODO: Ignore diagonal terms for better loss value. Currently these contribute very little
        pred[torch.eye(batch_size).byte()] = -1E9
        pred += self.C
        pred[targ == 0] *= -1
        logsigmoid = self.logsigmoid(pred)
        return -logsigmoid.sum() / n_pos

    def cross_entropy(self, pred):
        '''
        Args:
            pred: FloatTensor of inner product (plus bias) between representations of a batch
        Returns:
            unsupervised cross entropy loss
        '''

        batch_size = pred.shape[0]
        targ = self.target[:batch_size,:batch_size]
        pred[torch.eye(batch_size).byte()] = -1E9
        return torch.mean(torch.sum(-targ * self.logsoftmax(pred), 1))


    def forward(self, input):
        '''
        Args:
            input: FloatTensor or pair of FloatTensors (as a tuple) of a batch of representations
        Returns:
            unsupervised loss based on chosen objective function
        '''

        rep, context = input if type(input) == tuple else (input, input)
        if not self.transform is None:
            context = self.transform(context)
        pred = torch.matmul(rep, torch.t(context))
        if not self.bias is None:
            bias = self.bias(rep)
            pred += bias
            if type(input) == tuple or not self.transform is None:
                pred += self.bias(context).transpose(-1, 0)
            else:
                pred += bias.transpose(-1, 0)
        return self.objective(pred)


def obj_opts(parser):
    '''adds command-line options related to the objective
    Args:
        parser: argparse.ArgumentParser object
    Returns:
        updated parser
    '''

    parser.add_argument('--cross_entropy', action='store_true', help='use cross-entropy objective instead of NCE')
    parser.add_argument('--context_size', default=3, help='size of unsupervised context', type=int)
    parser.add_argument('--transform', default=None, help="context transform ('diagonal', 'linear', or 'psd')", type=str)
    parser.add_argument('--bias', default=None, help="functional bias term ('linear' or 'quadratic')", type=str)
    return parser

def init_obj(args):
    '''initializes objective from parsed arguments
    Args:
        args: parsed arguments
    Returns:
        UnFURL object
    '''

    kwargs = {'objective': 'cross_entropy' if args.cross_entropy else 'nce',
              'context_size': args.context_size,
              'transform': args.transform,
              'bias': args.bias}
    if args.verbose:
        write('Objective Settings:\n')
        print_dict(kwargs)
    criterion = UnFURL(max_batch_size=args.batch_size, dimension=args.output_dimension, **kwargs)
    if args.verbose:
        write("{}\n\n".format(criterion))
    return criterion
