import argparse
import pdb
import torch
from torch import nn
from text_embedding.documents import write
from UnFURL.nlp import Encoder
from UnFURL.nlp import TASKS
from UnFURL.nlp import TextEmbedding
from UnFURL.nlp import evaluate_model
from UnFURL.nlp import train_model
from UnFURL.objective import init_obj
from UnFURL.objective import obj_opts
from UnFURL.utils import cuda_device 
from UnFURL.utils import init_train
from UnFURL.utils import print_dict
from UnFURL.utils import train_opts


class GRUEncoder(Encoder):
    '''class for GRU sentence representations'''

    def __init__(self, input_dimension=300, output_dimension=1200, num_layers=1, bias=True, dropout=0.0, bidirectional=False, **kwargs):
        '''
        Args:
            input_dimension: word embedding dimension (also passed to Encoder)
            output_dimension: dimension of output representation
            num_layers: passed to nn.GRU
            bias: passed to nn.GRU
            dropout: passed to nn.GRU
            bidirectional: passed to nn.GRU
            kwargs: passed to Encoder
        '''

        super().__init__(input_dimension=input_dimension, **kwargs)
        hidden_dimension = int(output_dimension/2) if bidirectional else output_dimension
        self.encoder = nn.GRU(input_dimension, hidden_dimension, num_layers=num_layers, bias=bias, dropout=dropout if num_layers > 1 else 0.0, bidirectional=bidirectional)
        self.bidirectional = bidirectional
        self.dropout = nn.Dropout(dropout) if dropout else None

    def init_params(self):
        '''initialize model params
        '''

        for name, param in self.encoder.named_parameters():
            if 'bias' in name:
                nn.init.constant_(param, 1.0)
            elif 'weight' in name:
                nn.init.xavier_uniform_(param)
        return self

    def forward(self, input):
        '''
        Args:
            input: LongTensor of token ids
        Returns:
            FloatTensor of GRU sentence representations
        '''

        hidden = self.encoder(self.embedding(input).transpose(0, 1))[1]
        output = torch.cat((hidden[0], hidden[1]), 1) if self.bidirectional else hidden[0]
        return output if self.dropout is None else self.dropout(output)


def parse():

    parser = argparse.ArgumentParser(prog='python UnFURL/qt.py')
    parser.add_argument('corpus', help='corpus file')
    parser.add_argument('save', help='save file')
    parser.add_argument('-w', '--words', help='path to vocabulary file')
    parser.add_argument('-m', '--min_count', default=10, help='minimum word count in corpus', type=int)
    parser.add_argument('-l', '--max_length', default=30, help='max sentence length', type=int)
    parser.add_argument('-p', '--pretrained', help='path to pretained vectorfile')
    parser.add_argument('-i', '--input_dimension', default=300, help='embedding dimension', type=int)
    parser.add_argument('-o', '--output_dimension', default=1200, help='output embedding dimension', type=int)
    parser.add_argument('-d', '--dropout', default=0.0, help='dropout probability', type=float)
    parser.add_argument('-u', '--unidirectional', action='store_true', help='use uni-directional RNN')
    parser.add_argument('-c', '--context', action='store_true', help='train with separate contextual representation')
    obj_opts(parser)
    train_opts(parser)
    return parser.parse_args()

def main(args):

    verbose, debug = args.verbose, args.debug
    device = cuda_device(verbose=verbose)

    kwargs = {'context': args.context,
              'vectorfile': args.pretrained,
              'vocabfile': args.words,
              'min_count': args.min_count,
              'input_dimension': args.input_dimension,
              'output_dimension': args.output_dimension,
              'dropout': args.dropout,
              'bidirectional': not args.unidirectional,
              'max_length': args.max_length}
    if verbose: 
        write('Model Settings:\n')
        print_dict(kwargs)
    model = TextEmbedding(encoder=GRUEncoder, **kwargs)
    if verbose:
        write("{}\n\n".format(model))
    if debug:
        pdb.set_trace()

    criterion = init_obj(args)
    if debug:
        pdb.set_trace()

    kwargs = init_train(args)
    if debug:
        pdb.set_trace()
    if verbose:
        write('Training Model on '+args.corpus+'\n\n')
    train_model(model, criterion, args.corpus, device=device, **kwargs)
    torch.save(model.state_dict(), args.save)
    evaluate_model(model, tasks=TASKS, batch_size=args.batch_size, device=device)
    if args.debug:
        pdb.set_trace()


if __name__ == '__main__':

    main(parse())
