import numpy as np
import torch
from text_embedding.documents import write


FLOAT = np.float32


def cuda_device(verbose=False):
    '''gets available CUDA device
    Args:
        verbose: display output
    Returns:
        torch device
    '''

    if torch.cuda.is_available():
        if verbose:
            write('Using GPU\n')
        return torch.device('cuda')
    if verbose:
        write('Using CPU\n')
    return torch.device('cpu')


def print_dict(dictionary):
    '''prints dict items line by line
    Args:
        dictionary: Python dict
    Returns:
        number of items printed
    '''
        
    output = 0
    for k, v in dictionary.items():
        if not isinstance(v, dict):
            write("\t {}: {}\n".format(k, v))
            output += 1
    return output
    

def train_opts(parser, niter=1, batch_size=400, eta=5E-4, clip=0.0):
    '''adds command-line options related to training
    Args:
        parser: argparse.ArgumentParser object
        niter: default niter
        batch_size: default batch_size
        eta: default eta
        clip: default clip
    Returns:
        None
    '''

    parser.add_argument('-n', '--niter', default=niter, help='number of training epochs', type=int)
    parser.add_argument('-b', '--batch_size', default=batch_size, help='batch size', type=int)
    parser.add_argument('-e', '--eta', default=eta, help='learning rate', type=float)
    parser.add_argument('--clip', default=clip, help='norm of gradient clipping', type=float)
    parser.add_argument('-ef', '--eval_freq', default=0, help='evaluate model every few epochs', type=int)
    parser.add_argument('-sf', '--save_freq', default=0, help='save model every few epochs', type=int)
    parser.add_argument('-v', '--verbose', action='store_true', help='display output')
    parser.add_argument('-D', '--debug', action='store_true', help='run pdb')

def init_train(args):
    '''gets training kwargs
    Args:
        args: parsed arguments
    Returns:
        dict of training kwargs
    '''

    kwargs = {'epochs': args.niter,
              'batch_size': args.batch_size,
              'lr': args.eta,
              'clip': args.clip}
    if args.verbose:
        write('Training Settings:\n')
        print_dict(kwargs)
    kwargs.update({'eval_freq': args.eval_freq,
                   'save_freq': args.save_freq,
                   'save_loc': args.save,
                   'verbose': args.verbose,
                   'debug': args.debug})
    return kwargs
