import argparse
import numpy as np
from numpy.random import permutation
import os
from os.path import isfile, join
from sklearn.linear_model import LogisticRegression as Logit
from sklearn.metrics import zero_one_loss, log_loss
import sys
sys.path.append('../..')
import torch

from text_embedding.documents import *


WIKI_DIR = './Wiki-3029/'


def load_dataset(path, shuffle=True, dev=False, max_train=1E9):
    tr_labels, de_labels, te_labels = [], [], []
    tr_sents, de_sents, te_sents = [], [], []
    de = 0.1 * dev
    for i, fn in enumerate(os.listdir(path)):
        with open(join(path, fn), 'r') as f:
            data = [line.strip().lower() for line in f]
        n_tr = int((0.8-de)*len(data))
        n_de = int(de*len(data))
        n_te = len(data) - n_tr - n_de
        tr_sents.extend(data[:min(n_tr, max_train)])
        de_sents.extend(data[n_tr:n_tr+n_de])
        te_sents.extend(data[n_tr+n_de:])
        tr_labels += [i]*min(n_tr, max_train)
        de_labels += [i]*n_de
        te_labels += [i]*n_te
    if shuffle:
        idx = permutation(range(len(tr_sents)))
        tr_sents = [tr_sents[i] for i in idx]
        tr_labels = [tr_labels[i] for i in idx]
    if dev:
        return (tr_sents, tr_labels), (de_sents, de_labels), (te_sents, te_labels)
    return (tr_sents, tr_labels), (te_sents, te_labels)


def wiki3029(partitions=['train', 'test']):
    (tr_sents, tr_labels), (te_sents, te_labels) =\
        load_dataset(WIKI_DIR, shuffle=False)
    if len(partitions) == 2:
        return (tr_sents, tr_labels), (te_sents, te_labels)
    else:
        return (tr_sents, tr_labels)


def softmax(x):
    '''computes row-wise softmax
    Args:
        x: numpy 2d array
    Returns:
        numpy 2d array whose rows add up to 1
    '''
    return np.exp(x-x.max(1)[:,None])/np.exp(x-x.max(1)[:,None]).sum(1)[:,None]


def transform_(model, tensor, batch=100):
    '''runs the model on the input tensor
    Args:
        module: nn.Module for representation
        tensor: input tensor for the model
        batch: number of inputs passed to the model, for memory purposes
    Returns:
        torch.Tensor object of shape=(tensor.shape[0], out_dimension)
    '''
    if callable(getattr(model, 'eval', None)):
        model.eval()
    with torch.no_grad():
        return np.vstack([model(tensor[i:i+batch]).detach().cpu().numpy()
                              for i in range(0, len(tensor), batch)])
    if callable(getattr(model, 'train', None)):
        model.train()


def transform(model, sents, device=None):
    '''returns the output of the model for a list of sentences
    Args:
        module: nn.Module for representation
        sents: list of strings
    Returns:
        torch.Tensor object of shape=(len(sents), out_dimension)
    '''
    if device is None:
        device = next(model.parameters()).device
    with torch.no_grad():
        tensor = torch.cuda.LongTensor(
                     transform_(lambda x: model.str2long(x), sents)).to(device)
    return transform_(model, tensor)


def topk_acc(preds, labels, k=1, reduce=True):
    '''computes top-k accuracy of the predictions given the true labels
    Args:
        preds: numpy array of final layer
        labels: numpy array of the true labels
        reduce: if False, return an array of 0/1, else return the mean
    Returns:
        numpy array or bools or float
    '''
    if k == 1:
        correct = (preds.argmax(1) == labels)
    else:
        correct = np.any(np.argpartition(preds, -k, 1)[:,-k:] ==
                            labels[:,None], axis=1)
    if reduce:
        return correct.mean()
    else:
        return correct


def eval_multiclass(data, model, clf=None, finetune=False, k=5):
    '''evaluates the model on the labeled data
    Args:
        data: tuple of train and test data
        model: nn.Module for representation
        clf: nn.Linear to be used as classifier
        finetune: Not used
        k: also return top-k accuracy
    Returns:
        ((float, float), (float, float)):
        ((train top-1, train top-k), (test top-1, test top-k))
    '''
    model.eval()
    (tr_x_, tr_y_), (te_x_, te_y_) = data
    tr_y = tr_y_.detach().cpu().numpy()
    te_y = te_y_.detach().cpu().numpy()
    n_class = len(np.unique(tr_y))
    if clf == 'mean':
        tr_x = transform_(model, tr_x_)  
        te_x = transform_(model, te_x_)
        clf = np.vstack([tr_x[tr_y==i].mean(0) for i in range(n_class)])
        tr_preds = tr_x.dot(clf.T)
        te_preds = te_x.dot(clf.T)
    elif clf == 'learn':
        # Learn classifier
        return
    elif not clf is None:
        tr_preds = transform_(lambda tensor: clf(model(tensor)), tr_x_)
        te_preds = transform_(lambda tensor: clf(model(tensor)), te_x_)
    model.train()
    return (topk_acc(tr_preds, tr_y, 1), topk_acc(tr_preds, tr_y, k)),\
           (topk_acc(te_preds, te_y, 1), topk_acc(te_preds, te_y, k))


def avg_dist(src_vecs, tgt_vec, norm=1):
    return np.linalg.norm(src_vecs-tgt_vec[None,:], ord=norm, axis=1).mean()


def eval_acc(task, represent, norm=None):
    assert(task in TASKMAP['train-test split'])
    if norm is None:
        return evaluate(task, represent)
    else:
        return eval_metric_acc(task, represent, norm)


def eval_metric_acc(task, represent, norm):
    (dtr, ltr), (dte, lte) = TASKMAP['train-test split'][task]()
    tr_x = batched_build(dtr, represent, batchsize=100)
    te_x = batched_build(dte, represent, batchsize=100)
    tr_y = np.array(ltr)
    te_y = np.array(lte)
    return kwise_metric_acc(((tr_x, tr_y), (te_x, te_y)), norm)


def kwise_metric_acc(data, norm):
    (tr_x, tr_y), (te_x, te_y) = data
    classes = np.unique(tr_y)
    k = len(classes)
    tr_y = sum((tr_y==l)*i for i,l in enumerate(classes))
    te_y = sum((te_y==l)*i for i,l in enumerate(classes))
    tr_preds = np.array([np.argmin([avg_dist(tr_x[tr_y==y], vec, norm=norm) for y in range(k)]) for vec in tr_x])
    te_preds = np.array([np.argmin([avg_dist(tr_x[tr_y==y], vec, norm=norm) for y in range(k)]) for vec in te_x])
    tr_acc = 1 - zero_one_loss(tr_y, tr_preds)
    te_acc = 1 - zero_one_loss(te_y, te_preds)
    return 100*tr_acc, 100*te_acc


def kwise_lin_acc(data, model, clf=None):
    (tr_x, tr_y), (te_x, te_y) = data
    if not clf is None:
        weights = clf
        tr_prob = softmax(tr_x.dot(weights.T))
        te_prob = softmax(te_x.dot(weights.T))
        tr_loss_ = log_loss(tr_y, tr_prob) / np.log(2)
        te_loss_ = log_loss(te_y, te_prob) / np.log(2)
        tr_acc_ = 1-zero_one_loss(tr_y, tr_prob.argmax(1))
        te_acc_ = 1-zero_one_loss(te_y, te_prob.argmax(1))
    else:
        logit = Logit(fit_intercept=False, solver='liblinear',
                      multi_class='auto').fit(tr_x, tr_y)
        tr_loss_ = log_loss(tr_y, logit.predict_proba(tr_x)) / np.log(2)
        te_loss_ = log_loss(te_y, logit.predict_proba(te_x)) / np.log(2)
        tr_acc_ = logit.score(tr_x, tr_y)
        te_acc_ = logit.score(te_x, te_y)
    return (tr_acc_, te_acc_), (tr_loss_, te_loss_)


def avg_kwise_metric_acc(data, model, norm=1, k=2, n_task=100, seed=0):
    # TODO: separate function for acc of 1 task
    model.eval()
    (tr_x_, tr_y_), (te_x_, te_y_) = data
    tr_x = transform_(model, tr_x_)
    te_x = transform_(model, te_x_)
    tr_y = tr_y_.detach().cpu().numpy()
    te_y = te_y_.detach().cpu().numpy()

    classes = np.unique(tr_y)
    n_class = len(classes)
    class2idx = np.zeros(n_class)
    for i in range(n_class):
        class2idx[classes[i]] = i

    # Sample random k_wise classification tasks
    tr_acc, te_acc = 0, 0
    np.random.seed(seed)
    for n in range(n_task):
        k_classes = np.random.choice(classes, k, replace=False)
        tr_idx = np.any(tr_y[:,None]==k_classes[None,:], 1)
        te_idx = np.any(te_y[:,None]==k_classes[None,:], 1)
        tr_x_k, tr_y_k = tr_x[tr_idx], tr_y[tr_idx]
        te_x_k, te_y_k = te_x[te_idx], te_y[te_idx]
        tr_acc_, te_acc_ = kwise_metric_acc(((tr_x_k, tr_y_k), (te_x_k, te_y_k)), norm)
        tr_acc += tr_acc_
        te_acc += te_acc_
    model.train()
    return (tr_acc/n_task, te_acc/n_task)


def avg_kwise_lin_acc(data, model, clf=None, finetune=False, k=2, n_task=100, seed=0):
    '''evaluates the model on the average k-wise classification task by testing
       a many randomly sampled tasks
    Args:
        data: tuple of train and test data
        model: nn.Module for representation
        clf: nn.Linear to be used as classifier
        finetune: Not used
        k: number of classes in the classification problem 
        n_tests: number of supervised problems to test on
    Returns:
        (float, float): (train acc, test acc)
    '''
    # NOTE: Classes must be range(n) for some n
    model.eval()
    (tr_x_, tr_y_), (te_x_, te_y_) = data
    tr_x = transform_(model, tr_x_)
    te_x = transform_(model, te_x_)
    tr_y = tr_y_.detach().cpu().numpy()
    te_y = te_y_.detach().cpu().numpy()
    # Classes
    classes = np.unique(tr_y)
    class2idx = {c: i for i, c in enumerate(classes)}
    n_class = len(classes)
    # Prepare mean classifier if necessary
    if clf == 'mean':
        clf_np = np.vstack([tr_x[tr_y==y].mean(0) for y in classes])
    elif not clf is None:
        if not type(clf) is np.ndarray:
            clf_np = clf.weight.detach().cpu().numpy()
        else:
            clf_np = clf
    # Sample random k_wise classification tasks
    tr_acc, te_acc = 0, 0
    tr_loss, te_loss = 0, 0
    np.random.seed(seed)
    for n in range(n_task):
        k_classes = np.random.choice(classes, k, replace=False)
        tr_idx = np.any(tr_y[:,None]==k_classes[None,:], 1)
        te_idx = np.any(te_y[:,None]==k_classes[None,:], 1)
        tr_x_k, tr_y_k = tr_x[tr_idx], tr_y[tr_idx]
        te_x_k, te_y_k = te_x[te_idx], te_y[te_idx]
        tr_y_k = sum((tr_y_k==l)*i for i,l in enumerate(k_classes))
        te_y_k = sum((te_y_k==l)*i for i,l in enumerate(k_classes))
        if not clf is None:
            clf_k = clf_np[[class2idx[y] for y in k_classes]]
        else:
            clf_k = None
        (tr_acc_, te_acc_), (tr_loss_, te_loss_) =\
            kwise_lin_acc(((tr_x_k, tr_y_k), (te_x_k, te_y_k)), model, clf_k)
        tr_acc += tr_acc_
        te_acc += te_acc_
        tr_loss += tr_loss_
        te_loss += te_loss_
    model.train()
    return (tr_acc/n_task, te_acc/n_task), (tr_loss/n_task, te_loss/n_task)


def eval_sst(model):
    model.eval()
    task = 'sst'
    represent = lambda documents: transform(model, documents)
    evaluate(task, represent, prepare=None, invariant=True,
             verbose=False, intercept=False)
    model.train()


def parse():

    parser = argparse.ArgumentParser(prog='python multiclass.py', fromfile_prefix_chars='@')
    parser.add_argument('-ds', '--dataset', default=WIKI_DIR,
                        help='path to multiclass dataset')
    parser.add_argument('-c', '--cuda', default=0,
                        help='CUDA device number', type=int)
    # model arguments
    parser.add_argument('-w', '--words',
                        help='path to vocabulary file')
    parser.add_argument('-m', '--min_count', default=10,
                        help='minimum word count in corpus', type=int)
    parser.add_argument('-l', '--max_length', default=50,
                        help='max sentence length', type=int)
    parser.add_argument('-p', '--pretrained',
                        help='path to pretained vectorfile')
    parser.add_argument('-i', '--input_dimension', default=300,
                        help='embedding dimension', type=int)
    parser.add_argument('-o', '--output_dimension', default=300,
                        help='output embedding dimension', type=int)
    parser.add_argument('-d', '--dropout', default=0.2,
                        help='dropout probability', type=float)
    parser.add_argument('-r', '--norm_reg', default=0.0,
                        help='regularization for representation', type=float)
    parser.add_argument('-s', '--seed', default=-1,
                        help='torch seed, use positive integer', type=int)
    parser.add_argument('-g', '--gru', action='store_true',
                        help='use GRU instead of CBOW')
    parser.add_argument('-u', '--unidirectional', action='store_true',
                        help='use uni-directional RNN')
    parser.add_argument('-f', '--freeze', action='store_true',
                        help='won\'t train word embeddings')
    parser.add_argument('-me', '--meta', action='store_true',
                        help='meta learning setup')
    # training arguments
    parser.add_argument('-n', '--niter', default=1,
                        help='number of training epochs', type=int)
    parser.add_argument('-b', '--batch_size', default=50,
                        help='batch size', type=int)
    parser.add_argument('-e', '--eta', default=1E-3,
                        help='learning rate', type=float)
    parser.add_argument('-wd', '--weight_decay', default=0.0,
                        help='weight decay for regularization', type=float)
    parser.add_argument('-cl', '--clip', default=0.0,
                        help='clip gradient at this norm', type=float)
    parser.add_argument('-ef', '--eval_freq', default=0,
                        help='evaluate model every few epochs', type=int)
    parser.add_argument('-sf', '--save_freq', default=0,
                        help='save model every few epochs', type=int)
    parser.add_argument('-md', '--model_dir',
                        help='directory to save models in', type=str)
    parser.add_argument('-tc', '--train_clf', action='store_true',
                        help='updates parameters in the classifier')
    parser.add_argument('-sp', '--save_prefix',
                        help='path for saving models', type=str)
    parser.add_argument('-lp', '--load_prefix',
                        help='path for loading models', type=str)
    parser.add_argument('-v', '--verbose', action='store_true',
                        help='display output')
    return parser
