# python sup.py -n 40 -b 100 -e 0.0005 -m 10 -c 2 -tm -tc -g -d 0.2 -sp 'gru_b100_e0.0005_d0.2' > logs/sup/gru_b100_e0.0005_d0.2
# python sup.py -lp 'gru_b100_e0.001_d0.3_p'
import pdb
from torch.autograd import Variable
from torch import optim
from torch import nn

from multiclass import *
from UnFURL.nlp import Encoder
from UnFURL.qt import GRUEncoder
from unsup import unlabeled_data, unsup_loss


def sup_loss(model, clf, criterion, data, batch_size=100, k=10):
    '''computes supervised loss and top-1 and top-k accuracies of the model on
       the data
    Args:
        model: nn.Module for representation
        clf: nn.Module for classification
        criterion: loss function used
        data: tuple of data and labels
        batch_size: for memory reasons only
        k: int, also evaluate top-k accuracy
        sents: list of strings
    Returns:
        (float, float, float): (loss, top-1 acc, top-k acc)
    '''
    model.eval()
    x_, y_ = data
    loss = 0.0
    top1 = 0
    topk = 0
    nc = len(np.unique(y_))
    with torch.no_grad():
        for b in range(0, len(x_), batch_size):
            bx_ = x_[b:b+batch_size]
            by_ = y_[b:b+batch_size]
            pred = clf(model(bx_))
            pred_np = pred.detach().cpu().numpy()
            by_np = by_.detach().cpu().numpy()
            loss += len(bx_)*float(criterion(pred, by_))
            top1 += topk_acc(pred_np, by_np, 1, reduce=False).sum()
            topk += topk_acc(pred_np, by_np, k, reduce=False).sum()
    model.train()
    return loss/len(x_)/np.log(nc), top1/len(x_), topk/len(x_)


def train_sup_model(model, clf, data, train_model=True, train_clf=True,
                    epochs=1, opt=optim.Adam, batch_size=100, verbose=True,
                    device=torch.device('cuda'), eval_freq=0, save_prefix=None,
                    states=None, **kwargs):
    '''trains a model on supervised data
    Args:
        model: nn.Module for representation
        clf: nn.Module for classification
        data: tuple of training data and labels
        train_model: boolean, whether to train the parameters of the model
        train_clf: boolean, whether to train the classifier
        epochs: number of iteration of training
        opt: optimizer
        batch_size: for training
        verbose: boolean, whether to print
        device: cpu or gpu
        eval_freq: how often to evaluate
        save_prefix: location to save the state_dicts
        kwargs: passed to optimizer
    '''
    def trainable_params():
        if train_model:
            for param in model.parameters(): yield param
        if train_clf:
            for param in clf.parameters(): yield param

    def print_title():
        print('\repoch\t: (tr_l, de_l, te_l) | (tr_2_ac, tr_10_ac), (te_2_ac, te_10_ac) | (ul_tr, ul_tr_de, ul_all, ul_te)')
        sys.stdout.flush()

    def print_status(e):
        tr_l, tr_2, tr_10 = sup_loss(model, clf, criterion, (tr_x, tr_y))
        te_l, te_2, te_10 = sup_loss(model, clf, criterion, (te_x, te_y))
        de_l, de_2, de_10 = sup_loss(model, clf, criterion, (de_x, de_y))
        ul_tr = unsup_loss(model, blocks_tr, tr_x, criterion, 1, device) 
        ul_tr_de = unsup_loss(model, blocks_tr_de, tr_de_x, criterion, 1, device) 
        ul_all = unsup_loss(model, blocks_all, all_x, criterion, 1, device) 
        ul_te = unsup_loss(model, blocks_te, te_x, criterion, 1, device) 
        print('\r%d\t: (%.4f, %.4f, %.4f) | (%.4f, %.4f), (%.4f, %.4f) | (%.4f, %.4f, %.4f, %.4f)'%(
                    e, tr_l, de_l, te_l, tr_2, tr_10, te_2, te_10, ul_tr, ul_tr_de, ul_all, ul_te))
        sys.stdout.flush()
        return de_l

    # Set up model, optimizer and criteria
    model.to(device)
    clf.to(device)
    optimizer = opt(filter(lambda param: param.requires_grad,
                           trainable_params()),
                    **kwargs)
    criterion = nn.CrossEntropyLoss()
    (tr_x, tr_y), (te_x, te_y), (de_x, de_y) = data
    tr_de_x, tr_de_y = torch.cat([tr_x, de_x]), torch.cat([tr_y, de_y])
    all_x, all_y = torch.cat([tr_x, de_x, te_x]), torch.cat([tr_y, de_y, te_y])
    blocks_tr = unlabeled_data((tr_x, tr_y), 20, 2)
    blocks_tr_de = unlabeled_data((tr_de_x, tr_de_y), 20, 2)
    blocks_all = unlabeled_data((all_x, all_y), 20, 2)
    blocks_te = unlabeled_data((te_x, te_y), 20, 2)
    log_c = np.log(len(np.unique(tr_y)))

    # Load state if provided
    if states:
        assert(len(states) == 3)
        model.load_state_dict(states[0])
        clf.load_state_dict(states[1])
        optimizer.load_state_dict(states[2])

    # Start training
    best_val = 1E9
    best_state = None
    print_title()
    print_status(0)
    for epoch in range(epochs):
        total_loss = 0.0
        total_x = 0
        for b in range(0, len(tr_x), batch_size):
            batch_x = tr_x[b:b+batch_size]
            batch_y = tr_y[b:b+batch_size]
            batch_out = clf(model(batch_x))
            loss = criterion(batch_out, batch_y) / log_c
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()
            if verbose:
                total_loss += len(batch_x)*float(loss)
                total_x += len(batch_x)
                print('\r\t%d %d %.4f %.4f'%(epoch, b, float(loss), total_loss/total_x), end="")
        if eval_freq == 0 or (epoch+1) % eval_freq == 0:
            val = print_status(epoch+1)
            # if val > best_val:
            #     break
            best_val = val
            best_state = [model.state_dict().copy(),
                          clf.state_dict().copy(),
                          optimizer.state_dict().copy()]
    if eval_freq != 0 and (epoch+1) % eval_freq:
        print_status(epoch+1)
    if save_prefix:
        torch.save(best_state[0], 'models/sup/'+save_prefix+'_model')
        torch.save(best_state[1], 'models/sup/'+save_prefix+'_clf')
        torch.save(best_state[2], 'models/sup/'+save_prefix+'_optimizer')
    model.load_state_dict(best_state[0])
    clf.load_state_dict(best_state[1])


def main():
    parser = parse()
    sup_parse(parser)
    args = parser.parse_args()
    device = torch.device('cuda:'+str(args.cuda))
    print(args)
    sys.stdout.flush()

    # Store parser arguments to file
    if args.save_prefix:
        with open('args/sup/'+args.save_prefix, 'w') as f:
            f.write('\n'.join(sys.argv[1:]))

    # Load parser arguments from file
    load_prefix = args.load_prefix
    if load_prefix:
        args = parser.parse_args(['@args/sup/'+load_prefix])
        print(args)

    # Load data and shuffle training data
    if args.verbose:
        print('Loading data')
    (tr_sents, tr_labels), (de_sents, de_labels), (te_sents, te_labels) =\
        load_dataset(args.dataset, shuffle=True, dev=True, max_train=args.sup_data)

    # Model for representation
    if args.verbose:
        print('Initialize model')
    if args.gru:
        model = GRUEncoder(input_dimension=args.input_dimension,
                           output_dimension=args.output_dimension,
                           max_length=args.max_length, dropout=args.dropout,
                           freeze=args.freeze, vocabfile=args.words,
                           min_count=args.min_count, vectorfile=args.pretrained)
    else:
        model = Encoder(input_dimension=args.output_dimension,
                        max_length=args.max_length, freeze=args.freeze,
                        vocabfile=args.words, min_count=args.min_count,
                        vectorfile=args.pretrained)

    # Final classifier layer
    if args.verbose:
        print('Initialize classifiers')
    n_class = len(np.unique(tr_labels))
    clf = nn.Linear(args.output_dimension, n_class, bias=False)

    if args.verbose:
        print('Strings to LongTensors')
    tr_x = model.str2long(tr_sents, device=device)
    de_x = model.str2long(de_sents, device=device)
    te_x = model.str2long(te_sents, device=device)
    tr_y = Variable(torch.LongTensor(tr_labels)).to(device)
    de_y = Variable(torch.LongTensor(de_labels)).to(device)
    te_y = Variable(torch.LongTensor(te_labels)).to(device)
    data = ((tr_x, tr_y), (te_x, te_y), (de_x, de_y))
    if args.verbose:
        print(tr_x.shape, tr_y.shape, te_x.shape, te_y.shape)

    # Train
    opt_kwargs = {'lr': args.eta, 'weight_decay': args.weight_decay,
                  'epochs': args.niter}
    if load_prefix is None:
        if args.verbose:
            print('Train model')
        train_sup_model(model, clf, data,
                        train_model=args.train_model, train_clf=args.train_clf,
                        batch_size=args.batch_size, device=device,
                        verbose=args.verbose, eval_freq=args.eval_freq,
                        save_prefix=args.save_prefix, **opt_kwargs)
    else:
        model.load_state_dict(torch.load('models/sup/'+load_prefix+'_model'))
        model.to(device)
        clf.load_state_dict(torch.load('models/sup/'+load_prefix+'_clf'))
        clf.to(device)
    # print('multiclass clf (10)\t: ', eval_multiclass(data[:2], model, clf=clf, k=10))
    # print('multiclass mean (10)\t: ', eval_multiclass(data[:2], model, clf='mean', k=10))
    # print('avg_k_wise, clf (2)\t: ', avg_kwise_lin_acc(data[:2], model, clf=clf, k=2)[0])
    # print('avg_k_wise, mean (2)\t: ', avg_kwise_lin_acc(data[:2], model, clf='mean', k=2)[0])
    # print('avg_k_wise, None (2)\t: ', avg_kwise_lin_acc(data[:2], model, clf=None, k=2)[0])
    # print('avg_k_wise, clf (10)\t: ', avg_kwise_lin_acc(data[:2], model, clf=clf, k=10)[0])
    # print('avg_k_wise, mean (10)\t: ', avg_kwise_lin_acc(data[:2], model, clf='mean', k=10)[0])
    # print('avg_k_wise, None (10)\t: ', avg_kwise_lin_acc(data[:2], model, clf=None, k=10)[0])

    if load_prefix:
        pdb.set_trace()


def sup_parse(parser):
    parser.add_argument('-su', '--sup_data', default=1E9,
                        help='number of labeled examples used to train', type=int)
    parser.add_argument('-tm', '--train_model', action='store_true',
                        help='updates parameters in the model')


if __name__ == '__main__': main()
