# python unsup.py -n 10 -b 50 -v -e 0.01 -c 1 -k 1 -N 50 -K 2 -m 10
# python unsup.py -n 15 -b 50 -v -e 0.01 -c 1 -k 1 -N 20 -K 5 -m 10
import pdb
from torch.autograd import Variable
from torch import optim
from torch import nn

from multiclass import *
from UnFURL.nlp import Encoder
from UnFURL.qt import GRUEncoder


class Rep(nn.Module):

    def __init__(self, model, dimension, n_class, train=False):
        super().__init__()
        self.model = model
        self.fc = nn.Linear(dimension, n_class, bias=False)
        self.fc.weight.requires_grad = train
        self.sm = nn.Softmax()

    def str2long(self, input, **kwargs):
        return self.model.str2long(input, **kwargs)

    def forward(self, input):
        return self.sm(self.fc(self.model(input)))


def unlabeled_data(data, blocks_per_class, block_size=2, seed=0):
    '''generates unlabeled data as blocks
    Args:
        data: train data and labels
        blocks_per_class: number of unlabeled blocks
        block_size: number of samples per block
    Returns:
        numpy array of indices, shape=(num_classes*blocks_per_class, block_size)
    '''
    np.random.seed(seed)
    tr_x, tr_y = data
    tr_y = tr_y.detach().cpu().numpy()
    blocks = np.vstack([np.vstack([
                            np.random.choice(np.where(tr_y==i)[0], blocks_per_class)
                            for _ in range(block_size)]).T
                 for i in np.unique(tr_y)])
    np.random.shuffle(blocks)
    return blocks


def unsup_loss_batch(model, blocks, tr_x, criterion, pos_tensor, batch_size,
                     block_size, k, device, norm_reg=0.0,
                     pos='one_mean', neg='one'):

    log_k = np.log(k+1)
    loss = torch.Tensor([0.0]).to(device)
    if neg == 'one':
        neg_batch = np.random.choice(blocks.flatten(), k*batch_size)
        x2 = model(tr_x[neg_batch])
    if neg == 'mean' and not (pos == 'all_pairs'):
        neg_batch = blocks[np.random.choice(len(blocks), k*batch_size)]
        x2 = torch.stack([model(tr_x[neg_batch[:,i]])
                              for i in range(block_size)]).mean(0)

    if pos == 'one_mean' or pos == 'block_mean':
        indices = list(range(block_size))
        if pos == 'one_mean':
            indices = [0]
        for idx in indices:
            x = pos_tensor[idx]
            x1 = pos_tensor[np.array([i for i in range(block_size)
                                  if i!=idx])].mean(0)
            x1_x = (x1*x).sum(1)
            x2_x = torch.stack([(x2[i:i+batch_size]*x).sum(1)
                                    for i in range(0, x2.shape[0], batch_size)]).t()
            preds = torch.cat([x1_x.view(batch_size, 1), x2_x], 1)
            labels = torch.zeros(batch_size, dtype=torch.long).to(device)
            loss += criterion(preds, labels)
        loss /= len(indices)

    if pos == 'all_pairs':
        gen = ((x,y) for x in range(block_size) for y in range(block_size)
                  if x > y)
        for i, j in gen:
            x = pos_tensor[i]
            x1 = pos_tensor[j]
            x1_x = (x1*x).sum(1)
            x2_x = torch.stack([(x2[i:i+batch_size]*x).sum(1)
                                    for i in range(0, x2.shape[0], batch_size)]).t()
            preds = torch.cat([x1_x.view(batch_size, 1), x2_x], 1)
            labels = torch.zeros(batch_size, dtype=torch.long).to(device)
            loss += criterion(preds, labels)
        loss /= (block_size * (block_size-1) / 2)
    loss /= log_k
    if norm_reg > 0:
        loss += norm_reg*torch.norm(pos_tensor)**2/(block_size*pos_tensor.shape[1])
    return loss


def unsup_loss(model, blocks, tr_x, criterion, k, device,
               pos='one_mean', neg='one', batch_size=100):
    block_size = blocks.shape[1]
    model.eval()
    with torch.no_grad():
        total_loss = 0.0
        for b in range(0, len(blocks), batch_size):
            batch = blocks[b:b+batch_size]
            size = len(batch)
            batch_tensor = torch.stack([model(tr_x[batch[:,i]])
                                            for i in range(block_size)])
            loss = unsup_loss_batch(model, blocks, tr_x, criterion,
                                    batch_tensor, size, block_size, k,
                                    device, norm_reg=0.0, pos=pos, neg=neg)
            total_loss += size*float(loss)
        total_loss /= len(blocks)
    model.train()
    return total_loss
    


def train_unsup_model(model, u_data, s_data, k=1, block_size=2, pos='block',
                      neg='block', norm_reg=0.0, epochs=1, opt=optim.Adam,
                      batch_size=100, verbose=True, device=torch.device('cuda'),
                      eval_freq=0, save_prefix=None, save_freq=0,
                      model_dir=None, **kwargs):
    '''trains a model on unsupervised data
    Args:
        model: nn.Module for representation
        u_data: train and test unlabeled data
        s_data: train and test labeled data
        k: number of negative samples
        block_size: size of block containing inputs from the same class
        epochs: number of iteration of training
        opt: optimizer
        batch_size: for training
        verbose: boolean, whether to print
        device: cpu or gpu
        eval_freq: how often to evaluate
        kwargs: passed to optimizer
    '''
    def unsup_loss_(blocks_):
        return unsup_loss(model, blocks_, tr_x, criterion, k, device, pos, neg, batch_size)

    def print_title():
        print('\repoch\t: (tr_ul, te_ul) | (tr_2_sl, tr_k_sl),'
              '(te_2_sl, te_k_sl) | (tr_2_ac, tr_k_ac), (te_2_ac, te_k_ac)')
        sys.stdout.flush()

    def print_status(e):
        tr_ul= unsup_loss_(blocks)
        # de_ul= unsup_loss_(blocks_dev)
        te_ul = unsup_loss_(blocks_test)
        (tr_2, te_2), (tr_sl_2, te_sl_2) = avg_kwise_lin_acc(s_data, model, clf='mean', k=2)
        (tr_k, te_k), (tr_sl_k, te_sl_k) = avg_kwise_lin_acc(s_data, model, clf='mean', k=k+1)
        print('\r%d\t: (%.4f, %.4f) | (%.4f, %.4f), (%.4f, %.4f) | '
              '(%.4f, %.4f), (%.4f, %.4f) '%(e, tr_ul, te_ul, tr_sl_2,
                  tr_sl_k, te_sl_2, te_sl_k, tr_2, tr_k, te_2, te_k))
        sys.stdout.flush()
        return te_ul

    # Set up model, optimizer and criteria
    model.to(device)
    optimizer = opt(filter(lambda param: param.requires_grad,
                           model.parameters()),
                    **kwargs)
    criterion = nn.CrossEntropyLoss()

    # Unlabeled data
    (tr_x, _), (_, _) = s_data
    blocks, blocks_dev, blocks_test = u_data
    print(blocks.shape, blocks_test.shape)

    # Start training
    prev_val = 1E9
    prev_state = None
    best_found = False
    best_epoch = 0
    if save_prefix:
        if model_dir:
            prefix = model_dir + '/' + save_prefix
            if not os.path.isdir(model_dir):
                os.mkdir(model_dir)
        else:
            prefix = 'models/unsup/' + save_prefix

    print_title()
    print_status(0)
    for epoch in range(epochs):
        total_loss = 0.0
        total_blocks = 0
        np.random.shuffle(blocks)
        for b in range(0, len(blocks), batch_size):
            batch = blocks[b:b+batch_size]
            size = len(batch)
            batch_tensor = torch.stack([model(tr_x[batch[:,i]])
                                            for i in range(block_size)])
            loss = unsup_loss_batch(model, blocks, tr_x, criterion,
                                    batch_tensor, size, block_size, k, device,
                                    norm_reg=norm_reg, pos=pos, neg=neg)
            total_loss += size*float(loss)
            total_blocks += size
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()
            if verbose:
                print('\r\t%d %d %.4f %.4f'%\
                          (epoch, b, float(loss), total_loss/total_blocks),
                      end="")
        if eval_freq == 0 or (epoch+1) % eval_freq == 0:
            val = print_status(epoch+1)
            if not best_found and val > prev_val + 0.0001:
                best_found = True
                best_epoch = epoch
                if save_prefix:
                    torch.save(prev_state, prefix + '_' + str(epoch) + '_best_model')
            prev_val = val
            prev_state = model.state_dict().copy()
        if save_freq > 0 and (epoch+1) % save_freq == 0:
            torch.save(model.state_dict(), prefix + '_' + str(epoch+1) + '_model')
    if eval_freq != 0 and (epoch+1) % eval_freq:
        print_status(epoch+1)
    if save_freq > 0 and (epoch+1) % save_freq:
        torch.save(model.state_dict(), prefix + '_' + str(epoch+1) + '_model')
    return best_epoch


def main():
    parser = parse()
    unsup_parse(parser)
    args = parser.parse_args()
    device = torch.device('cuda:'+str(args.cuda))
    print(args)
    sys.stdout.flush()

    # Store parser arguments to file
    if args.save_prefix:
        with open('args/unsup/'+args.save_prefix, 'w') as f:
            f.write('\n'.join(sys.argv[1:]))

    # Load parser arguments from file
    load_prefix = args.load_prefix
    if load_prefix:
        args = parser.parse_args(['@args/unsup/'+load_prefix])
        print(args)

    # Load data and shuffle training data
    # print('Loading data')
    (tr_sents, tr_labels), (te_sents, te_labels) =\
        load_dataset(args.dataset, shuffle=False)

    # Model for representation
    # print('Initialize model')
    if args.seed >= 0:
        torch.manual_seed(args.seed)
    if args.gru:
        model = GRUEncoder(input_dimension=args.input_dimension,
                           output_dimension=args.output_dimension,
                           max_length=args.max_length, dropout=args.dropout,
                           freeze=args.freeze, vocabfile=args.words,
                           min_count=args.min_count, vectorfile=args.pretrained)
    else:
        model = Encoder(input_dimension=args.output_dimension,
                        max_length=args.max_length, freeze=args.freeze,
                        vocabfile=args.words, min_count=args.min_count,
                        vectorfile=args.pretrained)

    # Additional linear layer if required
    if args.num_classes:
        model = Rep(model, args.output_dimension, args.num_classes,
                    train=args.train_clf)
    model.to(device)

    # print('Labeled and Unlabeled data')
    tr_x = model.str2long(tr_sents, device=device)
    te_x = model.str2long(te_sents, device=device)
    tr_y = Variable(torch.LongTensor(tr_labels)).to(device)
    te_y = Variable(torch.LongTensor(te_labels)).to(device)
    if args.meta:
        np.random.seed(0)
        tr_classes = set(np.random.choice(np.unique(tr_y), 2500, replace=False))
        np.random.seed(None)
        tr_idx = np.array([int(int(y) in tr_classes) for y in tr_y])
        te_idx = np.array([int(int(y) in tr_classes) for y in te_y])
        print(tr_idx.sum(), (1-tr_idx).sum())
        tr_tr_x, tr_tr_y = tr_x[np.where(tr_idx)[0]], tr_y[np.where(tr_idx)[0]]
        tr_te_x, tr_te_y = te_x[np.where(te_idx)[0]], te_y[np.where(te_idx)[0]]
        te_tr_x, te_tr_y = tr_x[np.where(1-tr_idx)[0]], tr_y[np.where(1-tr_idx)[0]]
        te_te_x, te_te_y = te_x[np.where(1-te_idx)[0]], te_y[np.where(1-te_idx)[0]]
    else:
        tr_tr_x, tr_tr_y = tr_x, tr_y
        tr_te_x, tr_te_y = te_x, te_y
        te_tr_x, te_tr_y = tr_x, tr_y
        te_te_x, te_te_y = te_x, te_y
    print(tr_tr_x.shape, te_tr_x.shape)
    tr_s_data = ((tr_tr_x, tr_tr_y), (tr_te_x, tr_te_y))
    te_s_data = ((te_tr_x, te_tr_y), (te_te_x, te_te_y))
    # Unlabaled data
    blocks = unlabeled_data((tr_tr_x, tr_tr_y), args.blocks_per_class, args.block_size)
    blocks_dev = unlabeled_data((tr_tr_x, tr_tr_y), 20, args.block_size, seed=None)
    blocks_test = unlabeled_data((tr_tr_x, tr_tr_y), 20, args.block_size, seed=None)
    u_data = (blocks, blocks_dev, blocks_test)

    # Train
    # print('Train model')
    opt_kwargs = {'lr': args.eta, 'weight_decay': args.weight_decay}
    best = train_unsup_model(model, u_data, tr_s_data, k=args.neg_samples,
                      block_size=args.block_size, pos=args.pos, neg=args.neg,
                      norm_reg=args.norm_reg, epochs=args.niter, batch_size=args.batch_size,
                      eval_freq=args.eval_freq, verbose=args.verbose,
                      device=device, save_prefix=args.save_prefix,
                      save_freq=args.save_freq, model_dir=args.model_dir,
                      **opt_kwargs)
    # print('multiclass mean (10)\t: ', eval_multiclass(te_s_data, model, clf='mean', k=10))
    print('avg_k_wise, mean (2)\t: ', avg_kwise_lin_acc(tr_s_data, model, clf='mean', k=2)[0])
    print('avg_k_wise, None (2)\t: ', avg_kwise_lin_acc(tr_s_data, model, clf=None, k=2)[0])
    print('avg_k_wise, mean (2)\t: ', avg_kwise_lin_acc(te_s_data, model, clf='mean', k=2)[0])
    print('avg_k_wise, None (2)\t: ', avg_kwise_lin_acc(te_s_data, model, clf=None, k=2)[0])
    # print('avg_k_wise, mean (9)\t: ', avg_kwise_lin_acc(tr_s_data, model, clf='mean', k=9)[0])
    # print('avg_k_wise, None (9)\t: ', avg_kwise_lin_acc(tr_s_data, model, clf=None, k=9)[0])
    # print('avg_k_wise, mean (9)\t: ', avg_kwise_lin_acc(te_s_data, model, clf='mean', k=9)[0])
    # print('avg_k_wise, None (9)\t: ', avg_kwise_lin_acc(te_s_data, model, clf=None, k=9)[0])
    print('')
    pdb.set_trace()
    if args.save_prefix:
        best_state = torch.load(args.model_dir + '/' + args.save_prefix + '_%d_best_model'%best)
        if best_state:
            print('Best state: epoch ', best)
            model.load_state_dict(best_state)
        # print('multiclass mean (10)\t: ', eval_multiclass(te_s_data, model, clf='mean', k=10))
        print('avg_k_wise, mean (2)\t: ', avg_kwise_lin_acc(tr_s_data, model, clf='mean', k=2)[0])
        print('avg_k_wise, None (2)\t: ', avg_kwise_lin_acc(tr_s_data, model, clf=None, k=2)[0])
        print('avg_k_wise, mean (2)\t: ', avg_kwise_lin_acc(te_s_data, model, clf='mean', k=2)[0])
        print('avg_k_wise, None (2)\t: ', avg_kwise_lin_acc(te_s_data, model, clf=None, k=2)[0])
        # print('avg_k_wise, mean (10)\t: ', avg_kwise_lin_acc(te_s_data, model, clf='mean', k=9)[0])
        # print('avg_k_wise, None (10)\t: ', avg_kwise_lin_acc(te_s_data, model, clf=None, k=9)[0])
    # print(eval_multiclass(te_s_data, model, clf='mean', k=10))
    # print(avg_kwise_lin_acc(te_s_data, model, clf='mean', k=2))
    # print(avg_kwise_lin_acc(te_s_data, model, clf='mean', k=5))
    # print(avg_kwise_lin_acc(te_s_data, model, clf='mean', k=11))
    # torch.save(model.state_dict(), 'models/model1')


def unsup_parse(parser):
    parser.add_argument('-po', '--pos', default='one_mean',
                        help='algorithm for positive sample: one_mean, block_mean, all_pairs', type=str)
    parser.add_argument('-ne', '--neg', default='one',
                        help='algorithm for negative sample: one, mean', type=str)
    parser.add_argument('-k', '--neg_samples', default=1,
                        help='negative samples for unsup learning', type=int)
    parser.add_argument('-K', '--block_size', default=2,
                        help='number of samples per block', type=int)
    parser.add_argument('-N', '--blocks_per_class', default=20,
                        help='number of blocks per class', type=int)
    parser.add_argument('-C', '--num_classes',
                        help='number of classes for extra layer', type=int)


if __name__ == '__main__': main()
